# -*- coding: utf-8 -*-
from plone import api
from plone.app.imagecropping.browser.settings import ISettings
from plone.app.imagecropping.testing import IMAGECROPPING_INTEGRATION
from plone.app.testing import logout
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.registry.interfaces import IRegistry
from zope.component import getMultiAdapter
from zope.component import getUtility

import unittest

try:
    from Products.CMFPlone.utils import get_installer
except ImportError:
    # BBB for Plone 5.0 and lower.
    get_installer = None


class ControlPanelTestCase(unittest.TestCase):

    layer = IMAGECROPPING_INTEGRATION

    def setUp(self):
        self.portal = self.layer['portal']
        self.controlpanel = self.portal['portal_controlpanel']
        if get_installer is None:
            self.installer = api.portal.get_tool('portal_quickinstaller')
        else:
            self.installer = get_installer(self.portal)
        setRoles(self.portal, TEST_USER_ID, ['Manager'])

    def test_controlpanel_has_view(self):
        view = getMultiAdapter(
            (self.portal, self.portal.REQUEST),
            name='imagecropping-settings'
        )
        self.assertTrue(view())

    def test_controlpanel_view_is_protected(self):
        from AccessControl import Unauthorized
        logout()
        self.assertRaises(
            Unauthorized,
            self.portal.restrictedTraverse,
            '@@imagecropping-settings'
        )


class RegistryTestCase(unittest.TestCase):

    layer = IMAGECROPPING_INTEGRATION

    def setUp(self):
        self.portal = self.layer['portal']
        if get_installer is None:
            self.installer = api.portal.get_tool('portal_quickinstaller')
        else:
            self.installer = get_installer(self.portal)
        self.registry = getUtility(IRegistry)
        self.settings = self.registry.forInterface(ISettings)

    def test_records_removed_on_uninstall(self):
        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        if get_installer is None:
            self.installer.uninstallProducts(
                products=['plone.app.imagecropping'])
        else:
            self.installer.uninstall_product('plone.app.imagecropping')

        BASE_REGISTRY = \
            'plone.app.imagecropping.browser.settings.ISettings.{0:s}'
        records = (
            BASE_REGISTRY.format('large_size'),
            BASE_REGISTRY.format('min_size'),
        )

        for r in records:
            self.assertNotIn(r, self.registry)



# -*- coding: utf-8 -*-
from io import BytesIO
from plone.app.imagecropping.browser.settings import ISettings
from plone.app.imagecropping.interfaces import IImageCroppingMarker
from plone.app.imagecropping.interfaces import IImageCroppingUtils
from plone.app.imagecropping.storage import Storage
from plone.behavior.interfaces import IBehaviorAssignable
from plone.namedfile.interfaces import IImage
from plone.namedfile.interfaces import IImageScaleTraversable
from plone.namedfile.scaling import DefaultImageScalingFactory
from plone.registry.interfaces import IRegistry
from zope.component import adapter
from zope.component._api import getUtility
from zope.interface import implementer
from zope.schema import getFieldsInOrder

import PIL


class IImageCroppingDX(IImageScaleTraversable, IImageCroppingMarker):
    """Image cropping support marker interface for NamedFile/DX types
    """


@adapter(IImageCroppingDX)  # this would work almost also on AT!
class CroppingImageScalingFactory(DefaultImageScalingFactory):

    def _crop(self, data, box, default_format='PNG'):
        """crop data (image as open file) to box
        """
        image = PIL.Image.open(data)
        image_format = image.format or default_format
        cropped_image = image.crop(box)
        cropped_image_file = BytesIO()
        cropped_image.save(cropped_image_file, image_format, quality=100)
        cropped_image_file.seek(0)
        return cropped_image_file

    def create_scale(self, data, direction, height, width, **parameters):
        if self.box:
            # do crop stuff first
            data = self._crop(data, self.box)
        return super(CroppingImageScalingFactory, self).create_scale(
            data,
            direction,
            height,
            width,
            **parameters
        )

    def __call__(
        self,
        fieldname=None,
        direction='thumbnail',
        height=None,
        width=None,
        scale=None,
        **parameters
    ):
        storage = Storage(self.context)
        self.box = storage.read(fieldname, scale)
        if self.box:
            direction = 'down'
        else:
            registry = getUtility(IRegistry)
            settings = registry.forInterface(ISettings)
            if scale in settings.cropping_for:
                direction = 'down'
        return super(CroppingImageScalingFactory, self).__call__(
            fieldname=fieldname,
            direction=direction,
            height=height,
            width=width,
            scale=scale,
            **parameters
        )


@implementer(IImageCroppingUtils)
@adapter(IImageScaleTraversable)
class CroppingUtilsDexterity(object):
    """Dexterity variant of scaling adapter
    """

    def __init__(self, context):
        self.context = context

    def _all_fields(self):
        type_info = self.context.getTypeInfo()
        if type_info is None:
            return
        schema = type_info.lookupSchema()
        for field in getFieldsInOrder(schema):
            yield field
        behavior_assignable = IBehaviorAssignable(self.context)
        if behavior_assignable:
            for behavior in behavior_assignable.enumerateBehaviors():
                for field in getFieldsInOrder(behavior.interface):
                    yield field

    def _image_field_values(self):
        for fieldname, field in self._all_fields():
            value = getattr(self.context, fieldname, None)
            if value and IImage.providedBy(value):
                yield (fieldname, value)

    def image_fields(self):
        """ read interface
        """
        return [info[1] for info in self._image_field_values()]

    def image_field_names(self):
        """ read interface
        """
        return [info[0] for info in self._image_field_values()]

    def get_image_field(self, fieldname):
        """ read interface
        """
        return getattr(self.context, fieldname, None)

    def get_image_label(self, fieldname):
        for name, field in self._all_fields():
            if name == fieldname:
                return field.title
        return fieldname

    def get_image_data(self, fieldname):
        """ read interface
        """
        field = self.get_image_field(fieldname)
        return field.data

    def get_image_size(self, fieldname):
        """ read interface
        """
        field = self.get_image_field(fieldname)
        image_size = field.getImageSize()
        return image_size

    def save_cropped(self, fieldname, scale, image_file):
        """ see interface
        """
        # BBB - this is superseeded by usage of scaling factories
        pass



# -*- coding: utf-8 -*-
from io import BytesIO
from plone.app.imagecropping.dx import IImageCroppingDX
from plone.app.imagecropping.interfaces import IImageCroppingUtils
from plone.app.imagecropping.storage import Storage
from Products.Five.browser import BrowserView

import PIL.Image


class CroppingView(BrowserView):

    DEFAULT_FORMAT = 'PNG'

    def __call__(self, **kw):
        form = self.request.form
        fieldname = form['fieldname']
        scale_id = form['scale']
        if 'remove' in form:
            storage = Storage(self.context)
            storage.remove(fieldname, scale_id)
            return 'OK'

        box = (
            int(round(float(form['x']))),
            int(round(float(form['y']))),
            int(round(float(form['x']) + float(form['width']))),
            int(round(float(form['y']) + float(form['height']))),
        )
        self._crop(fieldname, scale_id, box)
        return 'OK'

    def _crop(self, fieldname, scale, box):
        """Delegate to store.

        """
        storage = Storage(self.context)
        storage.store(fieldname, scale, box)

        if IImageCroppingDX.providedBy(self.context):
            return
        # AT BBB scaling.
        croputils = IImageCroppingUtils(self.context)
        data = croputils.get_image_data(fieldname)

        original_file = BytesIO(data)
        image = PIL.Image.open(original_file)
        image_format = image.format or self.DEFAULT_FORMAT

        cropped_image = image.crop(box)
        cropped_image_file = BytesIO()
        cropped_image.save(cropped_image_file, image_format, quality=100)
        cropped_image_file.seek(0)

        croputils.save_cropped(fieldname, scale, cropped_image_file)



# -*- coding: utf-8 -*-
from operator import itemgetter
from plone import api
from plone.app.imagecropping.browser.settings import ISettings
from plone.app.imagecropping.interfaces import IImageCroppingUtils
from plone.app.imagecropping.storage import Storage
from plone.namedfile.interfaces import IAvailableSizes
from plone.registry.interfaces import IRegistry
from Products.Five.browser import BrowserView
from zope.component._api import getUtility
import six
from six.moves import map


class CroppingEditor(BrowserView):
    """ Cropping Editor View """

    def __init__(self, context, request):
        super(CroppingEditor, self).__init__(context, request)
        request.set('disable_plone.leftcolumn', 1)
        request.set('disable_plone.rightcolumn', 1)

    @property
    def _croputils(self):
        return IImageCroppingUtils(self.context)

    @property
    def _editor_settings(self):
        registry = getUtility(IRegistry)
        settings = registry.forInterface(ISettings)
        return settings

    @property
    def show_cropping(self):
        """returns True if there are any croppable scales on any of the fields
        """
        for fieldname in self.image_field_names:
            for value in self._scales(fieldname):
                return True
        return False

    def _min_size(self, image_size, scale_size):
        """ we need lower min-sizes if the image is smaller than the scale """
        width = scale_size[0]
        height = scale_size[1]
        if width > image_size[0]:
            ratio = float(image_size[0]) / float(width)
            width = image_size[0]
            height = float(height) * ratio
        if height > image_size[1]:
            ratio = float(image_size[1]) / float(height)
            height = image_size[1]
            width = float(width) * ratio
        return (int(round(width)), int(round(height)))

    def _initial_size(self, image_size, scale_size):
        """we need a best fit centered preselection to make editors life
        better.
        """
        ix, iy = list(map(float, image_size))

        # aspect ratio of original
        if iy > 0:
            ir = ix / iy
        else:
            ir = 1

        sx, sy = list(map(float, scale_size))

        # aspect ratio of scale
        if sy > 0:
            sr = sx / sy
        else:
            sr = 1

        # scale up to bounds
        if ir > sr:
            rx1, ry1 = ix * sr / ir, iy
        else:
            rx1, ry1 = ix, iy * ir / sr

        rx0, ry0 = 0, 0

        # center box
        if rx1 < ix:
            deltax = ix - rx1
            rx0 = deltax / 2
            rx1 = rx1 + deltax / 2
        if ry1 < iy:
            deltay = iy - ry1
            ry0 = deltay / 2
            ry1 = ry1 + deltay / 2

        # round to int
        rx0, ry0 = int(round(rx0)), int(round(ry0))
        rx1, ry1 = int(round(rx1)), int(round(ry1))
        return rx0, ry0, rx1, ry1

    def _scale_info(self, fieldname, scale_id, target_size, true_size):
        scale = dict()
        scale['id'] = scale_id
        scale['title'] = scale_id

        # scale config
        min_width, min_height = self._min_size(true_size, target_size)

        # lookup saved crop info
        storage = Storage(self.context)
        current_box = storage.read(fieldname, scale_id)

        if current_box is None:
            current_box = self._initial_size(true_size, target_size)
            scale['is_cropped'] = False
        else:
            scale['is_cropped'] = True

        # images original dimensions
        scale['true_width'] = true_size[0]
        scale['true_height'] = true_size[1]

        # images target dimensions
        scale['target_width'] = target_size[0]
        scale['target_height'] = target_size[1]

        # current selected crop
        scale['current'] = {
            'x': current_box[0],
            'y': current_box[1],
            'w': current_box[2] - current_box[0],
            'h': current_box[3] - current_box[1],
        }
        scale['aspect_ratio'] = '1.777778'  # 16:9
        if target_size[0] > 0 and target_size[1] > 0:
            scale['aspect_ratio'] = '{0:.2f}'.format(
                float(target_size[0]) / float(target_size[1])
            )
        scale['can_scale'] = (
            target_size[0] <= true_size[0] and
            target_size[1] <= true_size[1]
        )
        return scale

    def _scales(self, fieldname):
        constrain_cropping = self._editor_settings.constrain_cropping
        cropping_for = self._editor_settings.cropping_for
        allowed_sizes = getUtility(IAvailableSizes)() or []
        sizes_iterator = sorted(
            six.iteritems(allowed_sizes),
            key=itemgetter(1)
        )
        for scale_id, target_size in sizes_iterator:
            if constrain_cropping and scale_id not in cropping_for:
                continue
            yield scale_id, target_size

    def scales_info(self, fieldname):
        """Returns information to initialize pattern for all available scales
           on the current content with the given fieldname and interface."""
        true_size = self._croputils.get_image_size(fieldname)
        for scale_id, target_size in self._scales(fieldname):
            yield self._scale_info(
                fieldname,
                scale_id,
                target_size,
                true_size
            )

    @property
    def image_field_names(self):
        return self._croputils.image_field_names()

    def icon_url(self, fieldname):
        scales = api.content.get_view('images', self.context, self.request)
        scaled_img = scales.scale(
            fieldname,
            scale='icon',
            direction='keep',
        )
        return scaled_img and scaled_img.url or ''

    def original_url(self, fieldname):
        """Returns the url to the unscaled image"""
        url = self.context.absolute_url()
        return '{0}/@@images/{1}'.format(url, fieldname)

    def field_label(self, fieldname):
        return self._croputils.get_image_label(fieldname)



# -*- coding: utf-8 -*-
from persistent.dict import PersistentDict
from plone.app.imagecropping import PAI_STORAGE_KEY
from plone.app.imagecropping.events import CroppingInfoChangedEvent
from plone.app.imagecropping.events import CroppingInfoRemovedEvent
from plone.scale.storage import AnnotationStorage
from z3c.caching.purge import Purge
from zope.annotation.interfaces import IAnnotations
from zope.event import notify

import time


class Storage(object):

    def __init__(self, context):
        self.context = context

    def _key(self, fieldname, scale):
        return '{0:s}_{1:s}'.format(fieldname, scale)

    def _invalidate_scale(self, fieldname, scale):
        # Call storage with actual time in milliseconds.
        # This always invalidates old scales
        scale_storage = AnnotationStorage(
            self.context,
            int(time.time())
        )
        # holzhammermethode
        uids = list(scale_storage.keys())
        for uid in uids:
            del scale_storage[uid]

    def remove(self, fieldname, scale, surpress_event=False):
        # remove info from annotation
        key = self._key(fieldname, scale)
        if key in list(self._storage.keys()):
            del self._storage[key]
        self._invalidate_scale(fieldname, scale)
        if not surpress_event:
            notify(CroppingInfoRemovedEvent(self.context))
            notify(Purge(self.context))

    @property
    def _storage(self):
        return IAnnotations(
            self.context
        ).setdefault(
            PAI_STORAGE_KEY,
            PersistentDict()
        )

    def store(self, fieldname, scale, box):
        self.remove(fieldname, scale, surpress_event=True)
        key = self._key(fieldname, scale)
        self._storage[key] = box
        notify(CroppingInfoChangedEvent(self.context))

    def read(self, fieldname, scale):
        if not scale:
            return
        key = self._key(fieldname, scale)
        return self._storage.get(key)



# -*- coding: utf-8 -*-
from io import BytesIO
from os.path import dirname
from PIL import Image
from plone.namedfile.file import NamedBlobImage


TEST_IMAGE_FILE = '/'.join([dirname(__file__), u'plone-logo.png'])


def dummy_named_blob_png_image():
    return NamedBlobImage(
        data=open(TEST_IMAGE_FILE, 'rb').read(),
        filename=TEST_IMAGE_FILE
    )


def dummy_named_blob_jpg_image():
    img = Image.open(TEST_IMAGE_FILE)
    out = BytesIO()
    if img.mode in ('RGBA', 'LA'):
        # need to remove the alpha channel for Pillow > 4.1.x
        fill_color = '#FFFFFF'  # your background
        background = Image.new(img.mode[:-1], img.size, fill_color)
        background.paste(img, img.split()[-1])
        img = background
    img.save(out, format='JPEG', quality=75)
    out.seek(0)
    bi = NamedBlobImage(data=out.getvalue())
    out.close()
    return bi
